@@ -42,8 +42,10 @@
 
 #include <sys/time.h>
 #include <sys/resource.h>
+#include <sys/stat.h>
 #include <exception>
 #include <cstdio>
+#include <fcntl.h>
 #include <unistd.h>
 
 #include <oxt/initialize.hpp>
@@ -358,6 +360,27 @@ private:
 		return m_hasModXsendfile == YES;
 	}
 
+	static bool stderrEqualsFile(const char *path) {
+		struct stat s1, s2;
+
+		if (fstat(STDERR_FILENO, &s1) == -1) {
+			return false;
+		}
+
+		// No O_CREAT: we don't care if the file does not exist.
+		int fd = open(path, O_WRONLY | O_APPEND, 0600);
+		if (fd == -1) {
+			return false;
+		}
+		if (fstat(fd, &s2) == -1) {
+			close(fd);
+			return false;
+		}
+		close(fd);
+
+		return s1.st_dev == s2.st_dev && s1.st_ino == s2.st_ino && s1.st_rdev == s2.st_rdev;
+	}
+
 	int reportBusyException(request_rec *r) {
 		ap_custom_response(r, HTTP_SERVICE_UNAVAILABLE,
 			"This website is too busy right now.  Please try again later.");
@@ -1257,6 +1280,7 @@ public:
 			ok = LoggingKit::context->prepareConfigChange(loggingConfig,
 				errors, req);
 		} catch (const std::exception &e) {
+			ok = false;
 			fprintf(stderr, "ERROR: unable to configure logging system: %s\n", e.what());
 		}
 		if (ok) {
@@ -1346,6 +1370,8 @@ public:
 				" with an explicit log file using the `PassengerLogFile` directive.");
 		} else {
 			params.set("log_file", ap_server_root_relative(pconf, s->error_fname));
+			params.setBool("log_file_is_stderr", stderrEqualsFile(
+				ap_server_root_relative(pconf, s->error_fname)));
 		}
 
 		serverConfig.ctl.addTo(params);
@@ -1576,6 +1602,8 @@ destroy_hooks(void *arg) {
 		boost::this_thread::disable_interruption di;
 		boost::this_thread::disable_syscall_interruption dsi;
 		P_DEBUG("Shutting down Phusion Passenger...");
+		LoggingKit::shutdown();
+		oxt::shutdown();
 		delete hooks;
 		hooks = NULL;
 	} catch (const thread_interrupted &) {
@@ -1609,8 +1637,8 @@ init_module(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *
 	 */
 	if (hooks == NULL) {
 		oxt::initialize();
-		LoggingKit::initialize();
 		SystemTime::initialize();
+		LoggingKit::initialize();
 	} else {
 		P_DEBUG("Restarting Phusion Passenger....");
 		delete hooks;
